// Modified or written by Ex Machina SAGL for inclusion with lambdaj.
// Copyright (c) 2009 Mario Fusco.
// Licensed under the Apache License, Version 2.0 (the "License")

package ch.lambdaj.function.argument;

import java.lang.ref.*;
import java.lang.reflect.*;
import java.util.*;

import ch.lambdaj.util.*;

/**
 * Registers a method invocation
 * @author Mario Fusco
 */
final class Invocation {

	private final Class<?> invokedClass;
	private final Method invokedMethod;
	private String invokedPropertyName;
	private List<ParameterReference<Object>> weakArgs;

    Invocation(Class<?> invokedClass, Method invokedMethod, Object[] args) {
		this.invokedClass = invokedClass;
		this.invokedMethod = invokedMethod;
		invokedMethod.setAccessible(true);
		if (args != null && args.length > 0) {
			weakArgs = new LinkedList<ParameterReference<Object>>();
			for (int i = 0; i < args.length; i++) {
                weakArgs.add(new ParameterReference<Object>(args[i], !invokedMethod.getParameterTypes()[i].isPrimitive()));
            }
		}
	}
	
	private Object[] getConcreteArgs() {
		if (weakArgs == null) return new Object[0];
		Object[] args = new Object[weakArgs.size()];
		int i = 0;
		for (ParameterReference<Object> weakArg : weakArgs) { args[i++] = weakArg.get(); }
		return args;
	}

	Class<?> getInvokedClass() {
		return invokedClass;
	}

	Method getInvokedMethod() {
		return invokedMethod;
	}
	
	Class<?> getReturnType() {
		return invokedMethod.getReturnType();
	}
	
	String getInvokedPropertyName() {
		if (invokedPropertyName == null) invokedPropertyName = IntrospectionUtil.getPropertyName(invokedMethod);
		return invokedPropertyName;
	}

	Object invokeOn(Object object) throws InvocationException {
		try {
			return object == null ? null : invokedMethod.invoke(object, getConcreteArgs());
		} catch (Exception e) {
			throw new InvocationException(e, invokedMethod, object);
		}
	}
	
    /**
     * {@inheritDoc}
     */
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder(invokedMethod.toString());
		Object[] args = getConcreteArgs();
		if (args.length > 0) {
			sb.append(" with args ");
			for (int i = 0; i < args.length; i++) { sb.append(i == 0 ? "" : ", ").append(args[i]); }
		}
		return sb.toString();
	}
	
    /**
     * {@inheritDoc}
     */
	@Override
	public int hashCode() {
		int hashCode = 13 * invokedClass.hashCode();
		if (invokedMethod != null) hashCode += 17 * invokedMethod.hashCode();
		if (weakArgs != null) hashCode += 19 + weakArgs.size();
		return hashCode;
	}
	
    /**
     * {@inheritDoc}
     */
	@Override
	public boolean equals(Object object) {
		if (!(object instanceof Invocation)) return false;
		Invocation otherInvocation = (Invocation)object;
        return areNullSafeEquals(invokedClass, otherInvocation.getInvokedClass()) &&
               areNullSafeEquals(invokedMethod, otherInvocation.getInvokedMethod()) &&
               Arrays.equals(getConcreteArgs(), otherInvocation.getConcreteArgs());
	}
	
	private boolean areNullSafeEquals(Object first, Object second) {
		return (first == null && second == null) || (first != null && second != null && first.equals(second));
	}

    private static final class ParameterReference<T> {
        private final boolean garbageCollectable;
        private WeakReference<T> weakRef;
        private T strongRef;

        private ParameterReference(T referent, boolean garbageCollectable) {
            this.garbageCollectable = garbageCollectable;
            if (garbageCollectable) weakRef = new WeakReference<T>(referent);
            else strongRef = referent;
        }

        private T get() {
            return garbageCollectable ? weakRef.get() : strongRef;
        }
    }
}
